#!/usr/bin/perl -w

##
# @file gendocbook.pl: generate DocBook documentation from script comments
#
# @copyright 2002-2004: Axis Communications AB
# @author: Jonas Holmberg <jonashg@axis.com>
# 
# This script generates DocBook documentation from properly formatted comments
# of scripts.
#
# This comment will be converted to DocBook if this script is run with itself
# as argument. It recognises comment blocks starting with a line of just "##"
# that is followed by a line of either "# @file", "# @function" or "#
# @executable". Take a look at the examples below, and run
#
# @
# ./gendocbook.pl gendocbook.pl
# @
#
# to see what the output DocBook will look like.
#

##
# @function functionname: the purpose of the function in short
#
# @optional optionname: Description of an optional argument that may
#           continue on the next line like this.
# @optional $anotheroptionname: A replaceable optional argument. The name
#           "anotheroptionname" ought to be replaced with something else.
# @optional athirdoptionname=$replaceable: An optional argument that must be
#           followed by '=' and a replaceable name for the value it must be
#           assigned. The replaceable can be referred to in this description
#           as $replaceable.
# @optional afourthoptionname[=$replaceable]: An optional argument that may
#           optionally be followed by '=' and a name for the value it must be
#           assigned. The replaceable can be referred to in this description as
#           $replaceable.
# @optional afifthoptionname...: Description of yet another optional argument
#           that may be repeated. The other optionnames above can also be
#           followed by "..." to indicate repeativeness.
# @optional -l, --long-option: Description of an option that has both a short
#           and a long name.
# @required requiredargumentname
# @required $anotherrequiredargumentname
# @required athirdrequiredargumentname...
# @required $afourthrequiredargumentname...
#
# After an "empty" line (of just a '#') comes here a longer description of the
# function that may  consist of many lines until another line of just a '#'.
#
# This description should explain the required arguments, since they will not
# be included in the generated section about the options. They can be referred
# to as @requiredargumentname. @anotherrequiredargumentname and
# @afourthrequiredargumentname above are both replaceable. I.e. the names
# should be replaced by something when this function is called.
# @athirdrequiredargumentname may be repeated. A required argument can be both
# replaceable and repeated, like @afourthrequiredargumentname.
#
# The description will acually be presented before any descriptions of
# arguments and options in the generated documentation.
#
# This is yet another paragraph. Paragraphs are separated by empty lines.
#
# @The Title of a New Section:
#
# Following an empty line may a section be written that will be presented after
# the description of required arguments and options.
#
# Any number of sections consisting of a title and any number of paragraphs
# may be written.
#
# @Another Section:
#
# A paragraph that will be followed by a preformatted program listing that
# will be prosessed preserving the indentation and line breaks as long as it is
# surrounded by lines of only this string: "# @"
#
# @
# if [ this ]; then
#     that
# fi
# @
#
# @Other Tags:
#
# This is how you format a list:
#
# * Any word following a '@' will be interpreted as an option.
# * Any word following a '$' will be interpreted as a replaceable.
# * Any word following a '%' will be interpreted as a literal.
# * Any word followed by "()" will be interpreted as a function. A link to
#   that function will be generated as long as it's not the function being
#   described right here in this block of comments.
# * Any word followed by "*" will be interpreted as an executable. A link to
#   that execuatble will be generated as long as it's not the executable being
#   described right here in this block of comments.
# * Any file name preceeded by "file:" will be interpreted as a file name. If
#   the file name ends with a "/" it will be interpreted as a directory name.
#
# @Example:
#
# If my_echo* is given the option @-n or @--no-newline and the only following
# argument is $string, the output will be %42 without a following newline, as
# long as $string is %42. Like this:
#
# @
# my_echo -n 42
# @
#

##
# @executable my_echo: just an example
#
# @required $string
# @optional -n, --no-newline: Suppress newline.
#
# @Why This Executable?:
#
# So that the example above can be tested. And to show that you don't need to
# provide an untitled section (that will automatically be titled "Description")
# if you rather want your own title title for the first section in the
# generated documentation.
#



use strict;

#
# Print usage and exit.
#
sub usage {
        print STDERR "\nUsage: gendocbook.pl sourcefile [output-directory]\n\n";
        exit 1;
}

# File to be parsed
my $infile = shift @ARGV;
if (!$infile) {
	usage();
}
open INFILE, "<$infile" or
	die "Failed to open file \"$infile\" for reading";

# Output directory
my $Outdir = shift @ARGV;
if (!$Outdir) {
	$Outdir = ".";
}
-e $Outdir or system("mkdir -p $Outdir");

# Current line being parsed from infile
my $Line;	# Current line from infile being parsed.

# Parsing variables reset for each block of ducumentation
my $Name;	# Current block name
my $Purpose;	# Current purpose
my %Options;	# Current options
my @Arguments;	# Current arguments (required and options)
my %Sections;	# Current sections

# Parsing variables for collecting data
my @Titles;	# Current section titles
my %Functions;	# Purposes of all parsed functions
my $Entities;	# All system entities
my $FileName;	# Name of @file-tag

# Output indentation
my $In = " " x 4;	# Formatted text
my $PreIn = " " x 4;	# Preformatted text (programlisting)

# Initialize SGML variables
$Entities = "<!ENTITY all SYSTEM \"all.sgml\">\n";

#
# Indent text and break long lines.
#
sub indent(@) {
	(my $indent, my $text) = @_;
	my $linelen = 79;
	my $outline = $indent;
	my $ret = "";

	foreach my $inline (split(/\n/, $text)) {
		foreach my $word (split(/ /, $inline)) {
			if ($outline eq $indent or
			    length($outline) + length($word) + 1 <= $linelen) {
				$outline .= "$word ";
			} else {
				$ret .= "$outline\n";
				$outline = $indent . "$word ";
			}
		}
		if ($outline ne $indent) {
			$ret .= "$outline\n";
			$outline = $indent;
		}
	}
	if ($outline ne $indent) {
		$ret .= "$outline\n";
	}

	return $ret;
}

#
# Create a file and write contents.
#
sub write_file(@) {
	(my $filename, my $contents) = @_;

#$contents = "<!DOCTYPE refentry PUBLIC \"-//OASIS//DTD DocBook V4.1//EN\">\n$contents";
	open OUTFILE, ">$filename" or
		die "Failed to open file \"$filename\" for writing";
	print OUTFILE $contents or
		die "Failed to write to file \"$filename\"";
	print "Created file \"$filename\"\n";
}

my $preformatted = 0;
sub rest_of_para() {
	my $contents = "";

	while ((($Line = <INFILE>) =~ /^#\s+(.+?)\s*/) or $preformatted) {
		if ($Line =~ /^#\s+\@\w+.*:.*/ or
		    $Line =~ /^#\s+\@\w+\s+\$?\w+(\.\.\.)?\s*$/) {
			$preformatted = 0;
			last;
		}
		if ($Line =~ /^#\s+\@\s*$/) {
			$preformatted = 0;
			$contents .= "@"; # Newline will be added later
		} else {
			if ($preformatted) {
				$Line =~ s/^# ?(.*)$/$1/;
				$contents .= $Line;
			} else {
				$Line =~ s/^#\s+(.*?)\s*$/$1/;
				$contents .= " $Line";
			}
		}
	}

	return $contents;
}

#
# Read and discard one line from infile.
#
sub skip_line() {
	return $Line = <INFILE>
}

#
# Read and discard empty comments from infile.
#
sub skip_empty_lines() {
	my $ret = 0;

	while ($Line =~ /^#\s*$/) {
		$Line = <INFILE>;
		$ret = 1;
	}

	return $ret;
}

#
# Create refentry using global parsing variables and return SGML.
#
sub make_sgml(@) {
	(my $type) = @_;

	sub make_refsect1(@) {
		(my $title, my $section) = @_;

		my $sgml = "";

		my $id = "$Name-$title";
		$id =~ s/\s/_/g;
		$sgml .= "<refsect1 id=\"$id\">\n" .
			 "$In<title>$title</title>\n";
		my $preformatted = 0;
		foreach my $para (split(/\n/, $section)) {
			if ($para eq "@") {
				$preformatted = !$preformatted;
				if ($preformatted) {
					$sgml .= "$In<programlisting>\n";
				} else {
					$sgml .= "$In</programlisting>\n";
				}
			} else {
				if ($preformatted) {
					$sgml .= "$PreIn$para\n";
				} elsif ($para =~ s/\*\s+(.+)/$1/) {
					$sgml .= "$In<itemizedlist>\n";
					foreach my $item (split(/\* /, $para)) {
						$sgml .= "$In$In<listitem>\n";
						$sgml .= "$In$In$In<para>\n";
						$sgml .= indent("$In$In$In$In",
								$item);
						$sgml .= "$In$In$In</para>\n";
						$sgml .= "$In$In</listitem>\n";
					}
					$sgml .= "$In</itemizedlist>\n";
				} else {
					$sgml .= "$In<para>\n";
					$sgml .= indent("$In$In", $para);
					$sgml .= "$In</para>\n";
				}
			}
		}
		$sgml .= "</refsect1>\n\n";

		return $sgml;
	}

	sub process_tags(@) {
		(my $type, my $text) = @_;

		$text =~ s/<([\w.-]+)\@([\w.-]+)>/<email>$1&#64;$2<\/email>/g;
		$text =~ s/\@([\w-]+)/<option>$1<\/option>/g;
		$text =~ s/\$([\w\/]+)/<replaceable>$1<\/replaceable>/g;
		$text =~ s/\%([\w\/]+)/<literal>$1<\/literal>/g;
		$text =~ s/\bfile:([\w.\/-]*\/)([\s:\n])/
				"<filename class=\"directory\">$1<\/filename>$2"
			 /eg;
		$text =~ s/\bfile:([\w.\/-]+)\b/<filename>$1<\/filename>/g;
		$text =~ s/(\w+)\(\)/
			# Only create links to other functions.
			if (($type eq "function") and ($1 eq $Name)) {
				"<function>$1<\/function>";
			} else {
				"<link linkend=\"$1\"><citerefentry><refentrytitle><function>$1<\/function><\/refentrytitle><manvolnum>3<\/manvolnum><\/citerefentry><\/link>";
			}/eg;
		$text =~ s/([\w.]+)\*/
			# Only create links to other executables.
			if (($type eq "execuatble") and ($1 eq $Name)) {
				"<command>$1<\/command>";
			} else {
				"<link linkend=\"$1\"><citerefentry><refentrytitle><command>$1<\/command><\/refentrytitle><manvolnum>1<\/manvolnum><\/citerefentry><\/link>";
			}/eg;

		return $text;
	}

	foreach my $key (keys %Sections) {
		my $section = "";

		my $preformatted = 0;
		foreach my $para (split(/\n/, $Sections{$key})) {
			if ($para eq "@") {
				$preformatted = !$preformatted;
			} else {
				if (!$preformatted) {
					$para = process_tags($type, $para);
				}
			}
			$section .= "$para\n";
		}

		$Sections{$key} = $section;
	}

	my $tag;
	my $manvolnum;
	if ($type eq "file") {
		$tag = "filename";
		$manvolnum = "3";
	} elsif ($type eq "executable") {
		$tag = "command";
		$manvolnum = "1";
	} else {
		$tag = "function";
		$manvolnum = "3";
	}

	my $sgml;
	$sgml = "<refentry id=\"$Name\">\n\n" .
		"<refmeta>\n" .
		"$In<refentrytitle><$tag>$Name</$tag></refentrytitle>\n" .
		"$In<manvolnum>$manvolnum</manvolnum>\n" .
		"$In<refmiscinfo>Axis Build System</refmiscinfo>\n" .
		"</refmeta>\n\n" .
		"<refnamediv id=\"$Name--name\">\n" .
		"$In<refname>\n" .
		"$In$In<$tag>$Name</$tag>\n" .
		"$In</refname>\n" .
		"$In<refpurpose>\n" .
		"$In$In$Purpose\n" .
		"$In</refpurpose>\n" .
		"</refnamediv>\n\n";

	if ($type ne "file") {
		$sgml .= "<refsynopsisdiv id=\"$Name--synopsis\">\n" .
			 "$In<cmdsynopsis>\n" .
			 "$In$In<command>$Name</command>\n";

		foreach my $arg (@Arguments) {
			my $attr = "";
			if ($arg =~ /^(.+?)\.\.\.$/) {
				$arg = $1;
				$attr = " rep=\"repeat\"";
			}
			if ($arg =~ /^!(\$?[\w.]+)$/) {
				$attr .= " choice=\"req\"";
				$arg = $1;
				$arg =~ s/\$(.+)/
				     "<option><replaceable>$1<\/replaceable><\/option>"
				     /e;
			} elsif ($arg =~ /^([-\w\s,]+\w)\s*\[=?\$(\w+)\]/) {
				$arg = "<option>$1<optional>=" .
					"<replaceable>$2</replaceable>" .
					"</optional></option>";
			} elsif ($arg =~ /^([-\w]+)\s*=?\$(\w+)/) {
				$arg = "<option>$1=" .
				       "<replaceable>$2</replaceable></option>";
			} elsif ($arg =~ /^\$(\w+)/) {
				$arg = "<option><replaceable>$1</replaceable></option>";
			} else {
				$arg = "<option>$arg</option>";
			}
			$sgml .= "$In$In<arg$attr>$arg</arg>\n"
		}
		$sgml .= "$In</cmdsynopsis>\n" .
			 "</refsynopsisdiv>\n\n";
	}

	if ($Sections{"Description"}) {
		$sgml .= make_refsect1("Description", $Sections{"Description"});
	} else {
		if (!%Sections) {
			$Sections{"Description"} = "No comments.";
			@Titles = ( "Description" );
		}
	}

	if ($type eq "file") {
		$sgml .= "<refsect1 id=\"$Name--list\">\n" .
			 "$In<title>Functions</title>\n" .
			 "$In<para>\n" .
			 "$In$In" . "The following functions are documented " .
				"and available for use:\n" .
			 "$In</para>\n";
		$sgml .= "&$FileName.list;\n" .
			 "</refsect1>\n";

		$Entities .= "<!ENTITY $FileName.list SYSTEM " .
				"\"$FileName.list.sgml\">\n";
	} elsif (%Options) {
		$sgml .= "<refsect1 id=\"$Name-Options\">\n" .
			 "$In<title>Options</title>\n" .
			 "$In<variablelist>\n";
		# foreach my $key (sort keys %Options) {
		foreach my $key (sort {
					my $t1 = $a;
					my $t2 = $b;
					$t1 =~ s/^-*//;
					$t2 =~ s/^-*//;
					$t1 cmp $t2
				} keys %Options) {
			$Options{$key} =~ s/\@([-\w]+)/<option>$1<\/option>/g;
			$Options{$key} =~ s/\$([\w\/]+)/
				"<replaceable>$1<\/replaceable>"
				/eg;
			$Options{$key} =~ s/\%([\w\/]+)/
				"<literal>$1<\/literal>"
				/eg;

			my $option = $key;
			SWITCH: for ($option) {
			/([-\w\s,]+\w)(\s*)\[(=?)\$(\w*)\]/ and do {
				$option = "<option>$1$2<optional>$3" .
					  "<replaceable>$4</replaceable>" .
					  "</optional></option>";
				last;
			};
			/([-\w\s,]+\w)(\s*)(=?)\$(\w*)/ and do {
				$option = "<option>$1</option>$2$3" .
					  "<replaceable>$4</replaceable>";
				last;
			};
			/\$(\w+)/ and do { 
				$option = "<replaceable>$1</replaceable>";
				last;
			};
			/([-\w\s,]+\w)/ and do { 
				$option = "<option>$1</option>";
				last;
			};
			}

			$sgml .= "$In$In<varlistentry>\n" .
				 "$In$In$In<term>$option</term>\n" .
				 "$In$In$In<listitem>\n" .
				 "$In$In$In$In<para>\n";
			$sgml .= indent("$In$In$In$In$In",
					process_tags($type, $Options{$key}));
			$sgml .= "$In$In$In$In</para>\n" .
				 "$In$In$In</listitem>\n" .
				 "$In$In</varlistentry>\n";
		}
		$sgml .= "$In</variablelist>\n" .
			 "</refsect1>\n\n";
	}

	foreach my $title (@Titles) {
		$sgml .= make_refsect1($title, $Sections{$title});
	}

	$sgml .= "</refentry>\n";

	$Entities .= "<!ENTITY $Name SYSTEM \"$Name.sgml\">\n";

	return $sgml;
}

# Parse infile.
while ($Line = <INFILE>) {
	if (($Line =~ /##\s*\n/) and
	    (($Line = <INFILE>) =~ /^#\s+\@(\w+)[\s:]+([\w.]*):\s*(.*?)\s*$/)) {
		my $type =$1;
		my $author;
		my $copyright;
		$Name = $2;
		$Purpose = $3 . rest_of_para();

		print "TYPE: \"$type\"\n";
		print "NAME: \"$Name\"\n";
		print "PURPOSE: \"$Purpose\"\n";

		if ($type eq "function") {
			$Functions{$Name} = $Purpose;
		} elsif ($type eq "executable") {
			#$Executables{$Name} = $Purpose;
		}

		skip_empty_lines();

		if ($type eq "file") {
			if ($FileName) {
				die "Only one \@file allowed per file!";
			}
			$FileName = "$Name";
			while ($Line =~ /^#\s+\@(author)[\s:]+(.+?)\s*$/
			       or
			       $Line =~ /^#\s+\@(copyright)[\s:]+([0-9-]+):\s*(.*?)\s*$/) {
				if ($1 eq "author") {
					$author = $2 . rest_of_para();
					print "AUTHOR \"$author\"\n";
				} else {
					$copyright = "Copyright (C) $2  $3" .
						rest_of_para();
					print "COPYRIGHT \"$copyright\"\n";
				}
				skip_empty_lines();
			}
		} else {
			while ($Line =~ /^#\s+\@(required)[\s:]+(\$?\w+(\.\.\.)?)\s*/
			       or
			       $Line =~ /^#\s+\@(optional)[\s:]+(\$?[-\w\s,]+\w\s*(\[?=?\$?\w+\]?)?(\.\.\.)?):\s*(.*?)\s*$/) {
				if ($1 eq "required") {
					my $arg = $2 . rest_of_para();
					print "ARGUMENT \"$arg\"\n";
					push(@Arguments, "!$arg");
				} else {
					my $option = $2;
					push(@Arguments, $option);
					$Options{$option} = $5 . rest_of_para();
					print "OPTION \"$option\": " .
						"\"$Options{$option}\"\n";
				}
				skip_empty_lines();
			}
		}

		my $title;
		my $numtitles = 0;
		while ($Line =~ /^#/) {
			if (skip_empty_lines() and $Sections{$title}) {
				$Sections{$title} .= "\n";
			}
			if ($Line =~ /^#\s+\@(.*):\s*\n/s) {
				my $old_line = $Line;
				my $old_title = $title;
				skip_line();
				if (skip_empty_lines()) {
					$title = $1;
					$Titles[$numtitles++] = $title;
				} else {
					$Line =~ s/^#\s*(.*)\n/$1/;
					$Line = "$old_line\n" . $Line;
					$title = $old_title;
					if ($Sections{$title}) {
						$Sections{$title} .= "\n";
					}
				}
			} else {
				if (!$title) {
					$title = "Description";
				}
			}

			if ($Line =~ /^#/) {
				$Line =~ s/^#\s*(.*?)\s*$/$1/;
				if ($Line eq "@") {
					$preformatted = 1;
					$Sections{$title} .= "@\n";
				} else {
					$Sections{$title} .= $Line;
				}
				$Sections{$title} .= rest_of_para();
			}
		}

		if ($type eq "file") {
			$Sections{"Copyright"} = $copyright;
			$Titles[$numtitles++] = "Copyright";
			$Sections{"Author"} = $author;
			$Titles[$numtitles++] = "Author";
		}

		foreach my $key (keys %Sections) {
			print "SECTION $key: $Sections{$key}\n";
		}

		write_file("$Outdir/$Name.sgml", make_sgml($type));

		@Arguments = ();
		@Titles = ();
		foreach my $key (keys %Options) {
			delete $Options{$key};
		}
		foreach my $key (keys %Sections) {
			delete $Sections{$key};
		}
	}
}
close INFILE;

# Create list of all functions and their purpose in SGML.
my $FunctionList;
$FunctionList = "$In<variablelist>\n";
foreach my $key (sort(keys %Functions)) {
	$FunctionList .= "$In$In<varlistentry>\n" .
			 "$In$In$In<term><link linkend=\"$key\"><citerefentry><refentrytitle><function>$key</function></refentrytitle><manvolnum>3</manvolnum></citerefentry></link></term>\n" .
			 "$In$In$In<listitem>\n" .
			 "$In$In$In$In<para>\n";
	$FunctionList .= indent("$In$In$In$In$In", $Functions{$key});
	$FunctionList .= "$In$In$In$In</para>\n" .
			 "$In$In$In</listitem>\n" .
			 "$In$In</varlistentry>\n";
}
$FunctionList .= "$In</variablelist>\n";

# Create "global" files.
! $FileName and $FileName = "all";
write_file("$Outdir/$FileName.list.sgml", $FunctionList);
write_file("$Outdir/system.ent", $Entities);
write_file("$Outdir/all.sgml",
	   join("\n", map({$_="&$_;"} sort(keys %Functions))));

exit 0;
